"""Module to define impeller class."""
import csv
import multiprocessing
import warnings

import toml
from copy import deepcopy
from itertools import groupby
from pathlib import Path

import numpy as np
import plotly.graph_objects as go
from openpyxl import Workbook
from scipy.interpolate import interp1d, UnivariateSpline
from scipy.optimize import fsolve

from ccp import Q_, State, Point, Curve
from ccp.config.units import check_units
from ccp.config.utilities import r_getattr, r_setattr
from ccp.data_io.read_csv import read_data_from_engauge_csv


class ImpellerStateParameter:
    def __init__(self, impeller_state_object, attr):
        self.impeller_state_object = impeller_state_object
        self.attr = attr

    def __call__(self, *args, **kwargs):
        impeller_state_object = self.impeller_state_object
        attr = self.attr
        values = []

        for curve_state in impeller_state_object:
            values.append(getattr(getattr(curve_state, attr)(), "magnitude"))

        units = getattr(getattr(curve_state, attr)(), "units")

        return Q_(values, units)


def impeller_state_parameter(impeller_state_object, attr):
    return ImpellerStateParameter(impeller_state_object, attr)


class ImpellerState:
    def __init__(self, curves_state):
        self.curves_state = curves_state

        for attr in ["p", "T", "h", "s", "rho"]:
            func = impeller_state_parameter(self, attr)
            setattr(self, attr, func)

    def __getitem__(self, item):
        return self.curves_state.__getitem__(item)


class ImpellerPlotFunction:
    def __init__(self, impeller_object, attr):
        self.impeller_object = impeller_object
        self.attr = attr

    @check_units
    def __call__(self, *args, flow_v=None, speed=None, plot_kws=None, **kwargs):
        """Plot parameter versus volumetric flow.

        You can plot an specific point in the plot giving its flow_v and speed.

        You can choose units with the arguments flow_v_units='...' and
        {attr}_units='...'. For the speed you can use speed_units='...'.

        Parameters
        ----------
        flow_v : pint.Quantity, float, optional
            Volumetric flow (m³/s) for a specific point in the plot.
        speed : pint.Quantity, float, optional
            Speed (rad/s) for a specific point in the plot.
        flow_v_units : str, optional
            Flow units used for the plot. Default is m³/s.
        {attr}_units : str, optional
            Units for the parameter being plotted (e.g. for a head plot we could use
            head_units='J/kg' or head_units='J/g'. Default is SI.
        speed_units : str, optional
            Speed units for the plot. Default is 'rad/s'.

        Returns
        -------
        fig : plotly.Figure
            Plotly figure that can be customized.

        Examples
        --------
        >>> import ccp
        >>> imp = ccp.impeller_example()
        >>> fig = imp.plot_head(
        ...    flow_v=5.5,
        ...    speed=900,
        ...    flow_v_units='m³/h',
        ...    head_units='j/kg',
        ...    speed_units='RPM'
        ... )

        """
        impeller_object = self.impeller_object
        attr = self.attr
        fig = kwargs.pop("fig", None)

        if fig is None:
            fig = go.Figure()

        if plot_kws is None:
            plot_kws = {}

        p0 = impeller_object.points[0]
        flow_v_units = kwargs.get("flow_v_units", p0.flow_v.units)

        for curve in impeller_object.curves:
            fig = r_getattr(curve, attr + "_plot")(fig=fig, plot_kws=plot_kws, **kwargs)

        if speed:
            current_curve = impeller_object.curve(speed=speed)
            fig = r_getattr(current_curve, attr + "_plot")(
                fig=fig, plot_kws=plot_kws, **kwargs
            )
            if flow_v:
                current_point = impeller_object.point(flow_v=flow_v, speed=speed)
                fig = r_getattr(current_point, attr + "_plot")(
                    fig=fig, plot_kws=plot_kws, **kwargs
                )

        # extra x range
        flow_values = [p.flow_v.to(flow_v_units) for p in impeller_object.points]
        min_flow = min(flow_values)
        max_flow = max(flow_values)
        delta = 0.05 * (max_flow - min_flow)
        fig.update_layout(xaxis=dict(range=[min_flow - delta, max_flow + delta]))

        return fig


def impeller_plot_function(impeller_object, attr):
    return ImpellerPlotFunction(impeller_object, attr)


class Impeller:
    """An impeller with a performance map.

    Impeller instance is initialized with the list of points.
    The created instance will hold the dimensional points used in instantiation.
    Curves will be generated from points close in similarity.

    Parameters
    ----------
    points : list
        List with ccp.Point objects.

    Returns
    -------
    impeller : ccp.Impeller
    """

    @check_units
    def __init__(self, points):

        self.points = deepcopy(points)

        curves = []
        for speed, grouped_points in groupby(
            sorted(self.points, key=lambda point: point.speed),
            key=lambda point: point.speed,
        ):
            points = [point for point in grouped_points]
            curve = Curve(points)
            curves.append(curve)
            setattr(self, f"curve_{int(curve.speed.magnitude)}", curve)
        self.curves = curves
        self.disch = ImpellerState([c.disch for c in self.curves])

        for attr in [
            "disch.p",
            "disch.T",
            "disch.h",
            "disch.s",
            "disch.rho",
            "head",
            "eff",
            "power",
        ]:
            values = []
            # for disch.p etc values are defined in _Impeller_State
            if "." not in attr:
                for c in self.curves:
                    param = r_getattr(c, attr)
                    values.append(param.magnitude)
                units = param.units
                r_setattr(self, attr, Q_(values, units))

            r_setattr(self, attr + "_plot", impeller_plot_function(self, attr))

    def __getitem__(self, item):
        return self.points.__getitem__(item)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            points_other = sorted(other.points, key=lambda x: x.flow_v)
            points_self = sorted(self.points, key=lambda x: x.flow_v)
            if len(points_other) != len(points_self):
                return False
            else:
                return points_other == points_self

    @check_units
    def point(self, flow_v=None, flow_m=None, speed=None):
        """Calculate specific point in the performance map.

        Given a volumetric flow and a speed this method will calculate a point in the
        impeller map according to these arguments.

        Parameters
        ----------
        flow_v : pint.Quantity, float
            Volumetric flow (m³/s).
        flow_m : pint.Quantity, float
            Mass flow (kg/s).
        speed : pint.Quantity, float
            Speed (rad/s).

        Returns
        -------
        point : ccp.Point
            Point in the performance map.
        """

        current_curve = self.curve(speed)
        if flow_m:
            flow_v = current_curve.points[0].suc.v() * flow_m

        func_T = interp1d(
            current_curve.flow_v.m, current_curve.disch.T().m, fill_value="extrapolate"
        )
        func_p = interp1d(
            current_curve.flow_v.m, current_curve.disch.p().m, fill_value="extrapolate"
        )

        min_flow_v = min(current_curve.flow_v)
        max_flow_v = max(current_curve.flow_v)
        if flow_v < min_flow_v or max_flow_v < flow_v:
            warnings.warn(
                f"Expected point is being extrapolated.\n"
                f"Interpolation limits: {min_flow_v:.3f~P} ~ {max_flow_v:.3f~P}\n"
                f"Expected point flow: {flow_v:.3f~P}"
            )
        disch_T = func_T(flow_v)
        disch_p = func_p(flow_v)

        p0 = self.points[0]
        disch = State.define(p=disch_p, T=disch_T, fluid=p0.suc.fluid)

        point = Point(
            suc=p0.suc, disch=disch, flow_v=flow_v, speed=speed, b=p0.b, D=p0.D
        )

        return point

    @check_units
    def curve(self, speed=None):
        """Calculate specific point in the performance map.

        Given a speed this method will calculate a curve in the
        impeller map according to these arguments.

        Parameters
        ----------
        speed : pint.Quantity, float
            Speed (rad/s).

        Returns
        -------
        curve : ccp.Curve
            Point in the performance map.
        """
        speeds = np.array([curve.speed.magnitude for curve in self.curves])

        closest_curves_idxs = find_closest_speeds(speeds, speed.magnitude)
        curves = [
            self.curves[closest_curves_idxs[0]],
            self.curves[closest_curves_idxs[1]],
        ]

        # calculate factor
        speed_range = curves[1].speed.magnitude - curves[0].speed.magnitude
        factor = (speed.magnitude - curves[0].speed.magnitude) / speed_range

        current_curve = []
        p0 = self.points[0]
        number_of_points = len(curves[0])

        for i in range(number_of_points):
            flow_T, disch_T = get_interpolated_values(
                factor,
                curves[0].flow_v.magnitude[i],
                curves[0][i].disch.T().m,
                curves[1].flow_v.magnitude[i],
                curves[1][i].disch.T().m,
            )
            flow_p, disch_p = get_interpolated_values(
                factor,
                curves[0].flow_v.magnitude[i],
                curves[0][i].disch.p().m,
                curves[1].flow_v.magnitude[i],
                curves[1][i].disch.p().m,
            )

            disch = State.define(p=disch_p, T=disch_T, fluid=p0.suc.fluid)

            p = Point(
                suc=p0.suc,
                disch=disch,
                flow_v=(flow_T + flow_p) / 2,
                speed=speed,
                b=p0.b,
                D=p0.D,
            )

            current_curve.append(p)

        current_curve = Curve(current_curve)

        return current_curve

    @classmethod
    def convert_from(cls, original_impeller, suc=None, find="speed"):
        """Convert performance map from an impeller.

        Parameters
        ----------
        original_impeller : ccp.Impeller
            The original impeller.
        suc : ccp.State
            The new suction condition to which we want to convert to.
        find : str, optional
            The method in which the curves will be converted.
            For now only 'speed' is implemented, which means that, based on volume ratio,
            we calculate new values of speed for each curve and the respective discharge
            condition.

        Returns
        -------
        converted_impeller : ccp.Impeller
            The new impeller with the converted performance map for the required
            suction condition.
        """
        all_converted_points = []
        for curve in original_impeller.curves:
            with multiprocessing.Pool() as pool:
                converter_args = [(p, suc, find) for p in curve]
                converted_points = pool.map(converter, converter_args)

                speed_mean = np.mean([p.speed.magnitude for p in converted_points])

                converted_points = [
                    Point.convert_from(
                        p,
                        suc=p.suc,
                        find="volume_ratio",
                        speed=speed_mean,
                    )
                    for p in converted_points
                ]

                all_converted_points += converted_points

        return cls(all_converted_points)

    def _calc_new_points(self):
        """Calculate new dimensional points based on the suction condition."""
        #  keep volume ratio constant
        all_points = []
        for curve in self.curves:
            new_points = []
            for point in curve:
                new_points.append(self._calc_from_non_dimensional(point))

            speed_mean = np.mean([p.speed.magnitude for p in new_points])

            new_points = [
                self._calc_from_speed(p, Q_(speed_mean, p.speed.units))
                for p in new_points
            ]

            all_points += new_points

        self.new = self.__class__(
            all_points,
            b=self.b,
            D=self.D,
            _suc=self.new_suc,
            _flow_v=self.flow_v,
            _speed=self.speed,
        )

    def export_to_excel(self, path_name=None):
        """Export curves to excel file."""
        wb = Workbook()
        for curve in self.curves:
            sheet_name = f'{curve.speed.to("RPM"):.0f~P}'
            ws = wb.create_sheet(sheet_name)
            for i, p in enumerate(curve):
                i += 1  # openpyxl index
                if i == 1:
                    ws.cell(row=i, column=1, value="Flow (m**3/s)")
                    ws.cell(row=i, column=2, value="Head (kJ/kg)")
                    ws.cell(row=i, column=3, value="Efficiency (%)")
                else:
                    ws.cell(row=i, column=1, value=p.flow_v.magnitude)
                    ws.cell(row=i, column=2, value=p.head.to("kJ/kg").magnitude)
                    ws.cell(row=i, column=3, value=p.eff.magnitude * 100)

        if path_name is None:
            file_name = f'{self.suc.p().to("bar"):.0f~P}.xlsx'
            file_name = file_name.replace(" ", "-")
            path_name = Path.cwd() / file_name

        wb.save(str(path_name))

    @classmethod
    def load_from_dict(
        cls,
        suc,
        head_curves,
        eff_curves,
        b=None,
        D=None,
        number_of_points=10,
        flow_units="m**3/s",
        head_units="J/kg",
        speed_units="RPM",
        head_interpolation_method="interp1d",
        eff_interpolation_method="interp1d",
    ):
        """Create points from dict object.

        In this case the dict is in the following format:

        curve = {
            '10000': {
                'x': [list with flow values],
                'y': [list with head or eff values]
                },
            ...}
        Where 10000 is the speed for that curve.

        suc : ccp.State
            Suction state.
        head_curve : dict
            Dict with head/flow values.
        eff_curve : dict
            Dict with head/flow values.
        b : float, pint.Quantity
            Impeller width (m).
        D : float, pint.Quantity
            Impeller diameter (m).
        number_of_points : int
            Number of points that will be interpolated.
        flow_units : str
            Flow units used in the dict.
        head_units : str
            Head units used in the dict.
            If the curve head units are in meter you can use: head_units="m*g0".
        speed_units : str
            Speed units used in the dict.
        head_interpolation_method : str, optional
            Interpolation method from scipy. Can be interp1d or UnivariateSpline.
            Default is interp1d.
        eff_interpolation_method : str, optional
            Interpolation method from scipy. Can be interp1d or UnivariateSpline.
            Default is interp1d.
        """
        # define if we have volume or mass flow
        flow_type = "volumetric"
        if list(Q_(1, flow_units).dimensionality.keys())[0] == "[mass]":
            flow_type = "mass"

        points = []

        for speed, head_curve in head_curves.items():
            head_interpolated = interp1d(
                head_curve["x"],
                head_curve["y"],
                kind=3,
                fill_value="extrapolate",
            )
            head_spline = UnivariateSpline(
                head_curve["x"],
                head_curve["y"],
            )
            eff_curve = eff_curves[speed]
            # check eff scale
            if max(eff_curve["y"]) > 1:
                eff_curve["y"] = [i / 100 for i in eff_curve["y"]]
            eff_interpolated = interp1d(
                eff_curve["x"],
                eff_curve["y"],
                kind=3,
                fill_value="extrapolate",
            )
            eff_spline = UnivariateSpline(
                eff_curve["x"],
                eff_curve["y"],
            )

            # check for possible interpolation error by comparing interp1d and spline
            # avoid too much extrapolation
            min_x = max(head_curve["x"][0], eff_curve["x"][0])
            max_x = min(head_curve["x"][-1], eff_curve["x"][-1])
            points_x = np.linspace(min_x, max_x, number_of_points)

            head_max_error = max(
                1 - (head_interpolated(points_x) / head_spline(points_x))
            )
            eff_max_error = max(1 - (eff_interpolated(points_x) / eff_spline(points_x)))

            if head_max_error > 0.05:
                warnings.warn(
                    f"Head interpolation error in speed {speed} {speed_units}.\n"
                )
            if eff_max_error > 0.05:
                warnings.warn(
                    f"Efficiency interpolation error in speed {speed} {speed_units}.\n"
                )

            if head_interpolation_method == "UnivariateSpline":
                head_interpolated = head_spline
            if eff_interpolation_method == "UnivariateSpline":
                eff_interpolated = eff_spline

            points_head = head_interpolated(points_x)
            points_eff = eff_interpolated(points_x)

            args_list = [
                (
                    suc,
                    Q_(float(speed), speed_units),
                    Q_(flow, flow_units),
                    Q_(head, head_units),
                    eff,
                    b,
                    D,
                )
                for flow, head, eff in zip(points_x, points_head, points_eff)
            ]

            with multiprocessing.Pool() as pool:
                if flow_type == "volumetric":
                    points += pool.map(create_points_flow_v, args_list)
                else:
                    points += pool.map(create_points_flow_m, args_list)

        return cls(points)

    @classmethod
    def load_from_dict_isis(
        cls,
        suc,
        head_curves,
        eff_curves,
        b=None,
        D=None,
        number_of_points=10,
        flow_units="m**3/s",
        head_units="J/kg",
        speed_units="RPM",
    ):
        """Create points from dict object available in the ISIS platform.

        The dict is in the following format:
           head_curves_dict = [
        {
            "z": 11373,
            "points": [
                {"x": 94529, "y": 148.586},
                {"x": 98641, "y": 148.211},
                {"x": 101554, "y": 147.837},
            ...]
        ...,
        },
        ]

        Where z is the speed and each point is described as a dict with x and y
        pair where x is the flow and y is the head or eff.

        suc : ccp.State
            Suction state.
        head_curve : dict
            Dict with head/flow values.
        eff_curve : dict
            Dict with head/flow values.
        b : float, pint.Quantity
            Impeller width (m).
        D : float, pint.Quantity
            Impeller diameter (m).
        number_of_points : int
            Number of points that will be interpolated.
        flow_units : str
            Flow units used in the dict.
        head_units : str
            Head units used in the dict.
            If the curve head units are in meter you can use: head_units="m*g0".
        speed_units : str
            Speed units used in the dict.

        Examples
        --------
        >>> import ccp
        >>> composition_fd = dict(
        ...    n2=0.4,
        ...    co2=0.22,
        ...    methane=92.11,
        ...    ethane=4.94,
        ...    propane=1.71,
        ...    ibutane=0.24,
        ...    butane=0.3,
        ...    ipentane=0.04,
        ...    pentane=0.03,
        ...    hexane=0.01,
        ... )
        >>> suc_fd = State.define(p=Q_(3876, "kPa"), T=Q_(11, "degC"), fluid=composition_fd)
        >>> imp = ccp.Impeller.load_from_dict_isis(
        ...    suc=suc_fd,
        ...    head_curves=head_curves_dict,
        ...    eff_curves=eff_curves_dict,
        ...    b=Q_(10.6, "mm"),
        ...    D=Q_(390, "mm"),
        ...    number_of_points=6,
        ...    flow_units="kg/h",
        ...    head_units="kJ/kg",
        ... )
        """
        # change dict format from isis to that handled by the ccp method.

        head_curve_ccp = {}
        for head_curve in head_curves["CURVES"]:
            speed = str(head_curve["z"])
            x = []
            y = []
            for point in head_curve["points"]:
                x.append(point["x"])
                y.append(point["y"])

            head_curve_ccp[speed] = {}
            head_curve_ccp[speed]["x"] = x
            head_curve_ccp[speed]["y"] = y

        eff_curve_ccp = {}
        for eff_curve in eff_curves["CURVES"]:
            speed = str(eff_curve["z"])
            x = []
            y = []
            for point in eff_curve["points"]:
                x.append(point["x"])
                y.append(point["y"])

            eff_curve_ccp[speed] = {}
            eff_curve_ccp[speed]["x"] = x
            eff_curve_ccp[speed]["y"] = y

        return cls.load_from_dict(
            suc=suc,
            b=b,
            D=D,
            head_curves=head_curve_ccp,
            eff_curves=eff_curve_ccp,
            number_of_points=number_of_points,
            flow_units=flow_units,
            head_units=head_units,
            speed_units=speed_units,
        )

    @classmethod
    def load_from_engauge_csv(
        cls,
        suc,
        curve_name,
        curve_path,
        b=None,
        D=None,
        number_of_points=10,
        flow_units="m**3/s",
        head_units="J/kg",
        speed_units="RPM",
        head_interpolation_method="interp1d",
        eff_interpolation_method="interp1d",
    ):
        """Convert points from csv generated by engauge to csv with 6 points at same flow for use on hysys.

        The csv files should be generated with engauge with the following procedure:
        First, copy the image of the curve to your clipboard, then inside engauge digitizer:
            - Edit -> Paste as new
            - Name each curve with their respective speed value;
            - On Axis Point -> add 3 reference points
            - Select the curve (e.g. 10322 would be the curve for 10322 RPM)
            - Select the points using the segment fill tool;
            - Select the next curve and points...
            - Settings -> Export Setup -> Select:
            - Raws X's and Y's ; One curve on each line
        Files should be saved with the following convention:
            - <curve-name>-head.csv
            - <curve-name>-eff.csv

        suc : ccp.State
            Suction state.
        curve_path : pathlib.Path
            Path to the curves.
        curve_name : str
            Name for head and efficiency curve.
            Curves should have names <curve_name>-head.csv and <curve-name>-eff.csv.
        b : float, pint.Quantity
            Impeller width (m).
        D : float, pint.Quantity
            Impeller diameter (m).
        number_of_points : int
            Number of points that will be interpolated.
        flow_units : str
            Flow units used when extracting data with engauge.
        head_units : str
            Head units used when extracting data with engauge.
            If the curve head units are in meter you can use: head_units="m*g0".
        speed_units : str
            Speed units used when extracting data with engauge.
        head_interpolation_method : str, optional
            Interpolation method from scipy. Can be interp1d or UnivariateSpline.
            Default is interp1d.
        eff_interpolation_method : str, optional
            Interpolation method from scipy. Can be interp1d or UnivariateSpline.
            Default is interp1d.
        """
        head_path = curve_path / (curve_name + "-head.csv")
        eff_path = curve_path / (curve_name + "-eff.csv")

        head_curves = read_data_from_engauge_csv(head_path)
        eff_curves = read_data_from_engauge_csv(eff_path)

        return cls.load_from_dict(
            suc=suc,
            b=b,
            D=D,
            head_curves=head_curves,
            eff_curves=eff_curves,
            number_of_points=number_of_points,
            flow_units=flow_units,
            head_units=head_units,
            speed_units=speed_units,
            head_interpolation_method=head_interpolation_method,
            eff_interpolation_method=eff_interpolation_method,
        )

    def save(self, file):
        """Save impeller to a toml file.

        Parameters
        ----------
        file : str or pathlib.Path
            Filename to which the data is saved.
        """

        with open(file, mode="w") as f:
            # add points to file
            dict_to_save = {
                f"Point{i}": point._dict_to_save()
                for i, point in enumerate(self.points)
            }
            toml.dump(dict_to_save, f)

    @classmethod
    def load(cls, file):
        """Load impeller from toml file.

        Parameters
        ----------
        file : str or pathlib.Path
            Filename to which the data is saved.

        Returns
        -------
        impeller : ccp.Impeller
            Impeller object.
        """
        parameters = toml.load(file)
        points = [
            Point(**Point._dict_from_load(kwargs)) for kwargs in parameters.values()
        ]

        return cls(points)

    def save_hysys_csv(self, curve_dir):
        """Save curve to a csv with hysys format.

        curve_path: pathlib.Path
            Path for directory where the files will be saved.
        """
        curve_dir.mkdir(parents=True, exist_ok=True)
        surge = {"Speed (RPM)": [], "Volume Flow (m3/h)": []}
        stonewall = {"Speed (RPM)": [], "Volume Flow (m3/h)": []}

        for curve in self.curves:
            curve.save_hysys_csv(
                curve_dir / f'speed-{curve.speed.to("RPM").m:.0f}-RPM.csv'
            )
            surge["Speed (RPM)"].append(curve.speed.to("RPM").m)
            stonewall["Speed (RPM)"].append(curve.speed.to("RPM").m)
            surge["Volume Flow (m3/h)"].append(
                min(p.flow_v.to("m**3/h").m for p in curve)
            )
            stonewall["Volume Flow (m3/h)"].append(
                max(p.flow_v.to("m**3/h").m for p in curve)
            )

        for name, limit in zip(["surge", "stonewall"], [surge, stonewall]):
            with open(str(curve_dir / (name + ".csv")), mode="w") as csv_file:
                writer = csv.DictWriter(csv_file, fieldnames=list(limit.keys()))
                writer.writeheader()
                for speed, flow in zip(
                    limit["Speed (RPM)"], limit["Volume Flow (m3/h)"]
                ):
                    writer.writerow({"Speed (RPM)": speed, "Volume Flow (m3/h)": flow})


def find_closest_speeds(array, value):
    diff = array - value
    idx = np.abs(diff).argmin()

    if idx == 0:
        return [0, 1]
    elif idx == len(array) - 1:
        return [len(array) - 2, len(array) - 1]

    if diff[idx] > 0:
        idx = [idx - 1, idx]
    else:
        idx = [idx, idx + 1]

    return np.array(idx)


def get_interpolated_values(fac, flow_0, val_0, flow_1, val_1):
    x0 = [flow_0, val_0]
    if fac > 0.5:
        x0 = [flow_1, val_1]

    result = fsolve(system_to_interpolate, x0, args=(fac, flow_0, val_0, flow_1, val_1))

    flow_x = result[0]
    val_x = result[1]

    return flow_x, val_x


def system_to_interpolate(x, *args):
    fac, flow_0, val_0, flow_1, val_1 = args
    eq_1 = -x[1] + val_0 + (x[0] - flow_0) * (val_1 - val_0) / (flow_1 - flow_0)
    eq_2 = -fac * (flow_1 - flow_0) + x[0] - flow_0

    return [eq_1, eq_2]


def impeller_example():
    test_dir = Path(__file__).parent / "tests"
    data_dir = test_dir / "data"

    suc = State.define(
        p=Q_(4.08, "bar"),
        T=Q_(33.6, "degC"),
        fluid={
            "METHANE": 58.976,
            "ETHANE": 3.099,
            "PROPANE": 0.6,
            "N-BUTANE": 0.08,
            "I-BUTANE": 0.05,
            "N-PENTANE": 0.01,
            "I-PENTANE": 0.01,
            "NITROGEN": 0.55,
            "HYDROGEN SULFIDE": 0.02,
            "CARBON DIOXIDE": 36.605,
        },
    )
    imp = Impeller.load_from_engauge_csv(
        suc=suc,
        curve_name="lp-sec1-caso-a",
        curve_path=data_dir,
        b=Q_(5.7, "mm"),
        D=Q_(550, "mm"),
        head_units="kJ/kg",
        flow_units="m³/h",
        number_of_points=7,
    )

    return imp


def converter(x):
    """Helper function used to parallelize conversion of points."""
    point, suc, find = x
    return Point.convert_from(point, suc=suc, find=find)


def create_points_flow_v(x):
    """Helper function used to parallelize creation of points."""
    suc, speed, flow, head, eff, b, D = x
    return Point(
        suc=suc,
        speed=speed,
        flow_v=flow,
        head=head,
        eff=eff,
        b=b,
        D=D,
    )


def create_points_flow_m(x):
    """Helper function used to parallelize creation of points."""
    suc, speed, flow, head, eff, b, D = x
    return Point(
        suc=suc,
        speed=speed,
        flow_m=flow,
        head=head,
        eff=eff,
        b=b,
        D=D,
    )


def calc_min_head_point(x, speed, imp, min_head):
    """Helper function to calculate min_head point."""
    try:
        p = imp.point(flow_v=x, speed=speed)
        head = p.head.m
    # if state does not converge, force head value just to keep the iterations going
    except ValueError:
        head = -x + min_head
    return head - min_head



"""
ccp is a python library for calculation of centrifugal compressor performance.
It is based on the book of {cite}`ludtke2004process` and uses
[CoolProp](http://www.coolprop.org/) :cite:`coolprop` and
[REFPROP](https://www.nist.gov/srd/refprop) :cite:`refprop`
for the gas properties calculations.

```{code-block} python
import ccp

# ccp uses pint to handle units. Q_ is a pint quantity.
# If a pint quantity is not provided, SI units are assumed.
Q_ = ccp.Q_
ps = Q_(3, 'bar')
Ts = 300

# Define the fluid as a dictionary:
fluid = {
    "CarbonDioxide": 0.8,
    "Nitrogen": 0.2,
}

# Define suction and discharge states:

suc0 = ccp.State.define(fluid=fluid, p=ps, T=Ts)
disch0 = ccp.State.define(fluid=fluid, p=Q_(7.255, 'bar'), T=391.1)

# Create performance point(s):

point0 = ccp.Point(
    suc=suc0,
    disch=disch0,
    speed=Q_(7941, 'RPM'),
    flow_m=Q_(34203.6, 'kg/hr'),
    b=0.0285,
    D=0.365)
)
point1...

# Create an impeller with those points:

imp = Impeller([point0, point1, ...])
"""


###############################################################################
# set refprop path in the beginning to avoid strange behavior
###############################################################################

import os as _os
import warnings as _warnings
from pathlib import Path as _Path

import CoolProp.CoolProp as _CP
from ctREFPROP.ctREFPROP import REFPROPFunctionLibrary as _REFPROPFunctionLibrary

# use _ to avoid polluting the namespace when importing

try:
    _path = _Path(_os.environ["RPPREFIX"])
except KeyError:
    if _os.path.exists("C:\\Users\\Public\\REFPROP"):
        _os.environ["RPprefix"] = "C:\\Users\\Public\\REFPROP"
        _path = _Path(_os.environ["RPPREFIX"])
    else:
        _path = _Path.cwd()

_CP.set_config_string(_CP.ALTERNATIVE_REFPROP_PATH, str(_path))
try:
    _RP = _REFPROPFunctionLibrary(_path)
    _RP.SETPATHdll(str(_path))
except TypeError:
    _RP = _REFPROPFunctionLibrary

if _os.name == "posix":
    _shared_library = "librefprop.so"
else:
    _shared_library = "REFPRP64.DLL"

_library_path = _path / _shared_library

if not _library_path.is_file():
    _warnings.warn(f"{_library_path}.\nREFPROP not configured.")

__version__ = "0.1.18rc2"

__version__full = (
    f"ccp: {__version__} | "
    + f'CP : {_CP.get_global_param_string("version")} | '
    + f'REFPROP : {_CP.get_global_param_string("REFPROP_version")}'
)

###############################################################################
# pint
###############################################################################
import pint as _pint

_new_units = _Path(__file__).parent / "config/new_units.txt"
ureg = _pint.get_application_registry()
if isinstance(ureg.get(), _pint.registry.LazyRegistry):
    ureg = _pint.UnitRegistry()
    ureg.load_definitions(str(_new_units))
    # set ureg to make pickle possible
    _pint.set_application_registry(ureg)

Q_ = ureg.Quantity
_warnings.filterwarnings("ignore", message="The unit of the quantity is stripped")

###############################################################################
# plotly theme
###############################################################################

from plotly import io as _pio
import ccp.plotly_theme

_pio.templates.default = "ccp"

###############################################################################
# imports
###############################################################################

from .config.fluids import fluid_list
from .state import State
from .point import Point
from .curve import Curve
from .impeller import Impeller, impeller_example
from .fo import FlowOrifice
from .similarity import check_similarity

__all__ = [
    "State",
    "Point",
    "Curve",
    "Impeller",
    "FlowOrifice",
    "fluid_list",
    "check_similarity",
    "impeller_example",
]



"""Make a cProfile from main ccp functions.

To use it:
$ python cprofile.py <function name>

This will create a .prof file that can be visualized with snakeviz
$ pip install snakeviz
$ snakeviz <prof file name>.prof

"""
import cProfile
import subprocess
import sys
from pathlib import Path

from ccp import Q_, State, Point, Impeller


def generate_label(caller):
    """Generates a label to the cprofile output file."""
    commit_number = subprocess.check_output(["git", "describe", "--always"])
    commit_number = str(commit_number, "utf-8").strip("\n")

    increment = 0
    label = Path(caller + "_" + commit_number + "_" + str(increment) + ".prof")

    while label.exists():
        increment += 1
        label = Path(caller + "_" + commit_number + "_" + str(increment) + ".prof")

    return label


def state():
    State.define(
        rho=0.9280595769591103, p=Q_(1, "bar"), fluid={"Methane": 0.5, "Ethane": 0.5}
    )


def impeller():
    fluid = dict(
        methane=0.69945,
        ethane=0.09729,
        propane=0.0557,
        nbutane=0.0178,
        ibutane=0.0102,
        npentane=0.0039,
        ipentane=0.0036,
        nhexane=0.0018,
        n2=0.0149,
        co2=0.09259,
        h2s=0.00017,
        water=0.002,
    )
    suc = State.define(p=Q_(1.6995, "MPa"), T=311.55, fluid=fluid)

    p0 = Point(
        suc=suc,
        flow_v=Q_(6501.67, "m**3/h"),
        speed=Q_(11145, "RPM"),
        head=Q_(179.275, "kJ/kg"),
        eff=0.826357,
        b=Q_(28.5, "mm"),
        D=Q_(365, "mm"),
    )
    p1 = Point(
        suc=suc,
        flow_v=Q_(7016.72, "m**3/h"),
        speed=Q_(11145, "RPM"),
        head=Q_(173.057, "kJ/kg"),
        eff=0.834625,
        b=Q_(28.5, "mm"),
        D=Q_(365, "mm"),
    )

    imp1 = Impeller(
        [p0, p1],
    )


def create_ccp_points():
    composition_fd = dict(
        n2=0.4,
        co2=0.22,
        methane=92.11,
        ethane=4.94,
        propane=1.71,
        ibutane=0.24,
        butane=0.3,
        ipentane=0.04,
        pentane=0.03,
        hexane=0.01,
    )
    suc_fd = State.define(p=Q_(3876, "kPa"), T=Q_(11, "degC"), fluid=composition_fd)

    test_dir = Path.home() / "ccp/ccp/tests"
    curve_path = test_dir / "data"
    curve_name = "normal"

    imp_fd = Impeller.load_from_engauge_csv(
        suc=suc_fd,
        curve_name=curve_name,
        curve_path=curve_path,
        b=Q_(10.6, "mm"),
        D=Q_(390, "mm"),
        number_of_points=6,
        flow_units="kg/hr",
        head_units="kJ/kg",
    )
    new_fluid = {"co2": 0.7, "n2": 0.3}
    new_suc = State.define(p=Q_(10, "bar"), T=Q_(40, "degC"), fluid=new_fluid)
    imp_conv = Impeller.convert_from(imp_fd, suc=new_suc, find="speed")


if __name__ == "__main__":
    func = sys.argv[1]
    file_name = generate_label(func)
    cProfile.run(func + "()", file_name)



import pytest
import ccp
from numpy.testing import assert_allclose

Q_ = ccp.Q_


@pytest.fixture
def fo0():
    n2 = ccp.State.define(
        p=Q_(9.85, "bar"), T=Q_(145, "degC"), fluid={"n2": 1 - 1e-15, "o2": 1e-15}
    )
    qm = Q_(71281, "kg/hr")
    delta_p = Q_(9.85 - 4, "bar")
    D = Q_(0.3032, "m")

    return ccp.FlowOrifice(n2, delta_p, qm, D)


def test_flow_orifice(fo0):
    assert_allclose(fo0.d.magnitude, 0.10960820617559523)
    fo0.D = 0.5
    fo0.update()
    assert_allclose(fo0.d.magnitude, 0.11028294814259916)



"""Flow orifice."""
import numpy as np
from scipy.optimize import newton

from . import Q_


class FlowOrifice:
    def __init__(self, state=None, delta_p=None, qm=None, D=None, tappings="flange"):
        self.state = state
        self.delta_p = delta_p
        self.qm = qm
        self.D = D.to("m")
        self.tappings = tappings

        d = newton(
            calc_d, 0.1, args=(self.state, self.delta_p, self.qm, self.D, self.tappings)
        )
        self.d = Q_(float(d), "m")
        _, self.beta, self.C, self.e, self.reynolds, self.A = calc(
            self.d.magnitude, self.state, self.delta_p, self.qm, self.D, self.tappings
        )

    def update(self):
        d = newton(
            calc_d, 0.1, args=(self.state, self.delta_p, self.qm, self.D, self.tappings)
        )
        self.d = Q_(float(d), "m")

    def __setattr__(self, name, value):
        try:
            if isinstance(getattr(self, name), Q_) and not isinstance(value, Q_):
                super().__setattr__(name, Q_(value, getattr(self, name).units))
            else:
                super().__setattr__(name, value)

        except AttributeError:
            super().__setattr__(name, value)


def calc(x, state=None, delta_p=None, qm=None, D=None, tappings="flange"):
    d = Q_(x, "m")

    # calc beta
    beta = d / D

    p1 = state.p()
    p2 = p1 + delta_p
    rho = state.rho()
    A = np.pi * D**2 / 4
    u = qm / (rho * A)
    reyn = (rho * u * D / state.viscosity()).to("dimensionless")

    if tappings == "corner":
        L1 = L2 = 0
    elif tappings == "D D/2":
        L1 = 1
        L2 = 0.47
    elif tappings == "flange":
        L1 = L2 = Q_(0.0254, "m") / D
    else:
        raise ValueError('tappings must be "corner", "D D/2" or "flange"')

    M2 = 2 * L2 / (1 - beta)

    # calc C
    C = (
        0.5961
        + 0.0261 * beta**2
        - 0.216 * beta**8
        + 0.000521 * (1e6 * beta / reyn) ** 0.7
        + (0.0188 + 0.0063 * (19000 * beta / reyn) ** 0.8)
        * beta**3.5
        * (1e6 / reyn) ** 0.3
        + (0.043 + 0.080 * np.e ** (-10 * L1) - 0.123 * np.e ** (-7 * L1))
        * (1 - 0.11 * (19000 * beta / reyn) ** 0.8)
        * (beta**4 / (1 - beta**4))
        - 0.031 * (M2 - 0.8 * M2**1.1) * beta**1.3
    )

    if D < Q_(71.12, "mm"):
        C += 0.011 * (0.75 - beta) * (2.8 - D / 0.0254)

    # calc e
    e = 1 - (0.351 + 0.256 * beta**4 + 0.93 * beta**8) * (
        1 - (p2 / p1) ** (1 / state.kT())
    )

    # calc qm
    qm_calc = (
        C
        / (np.sqrt(1 - beta**4))
        * e
        * (np.pi / 4)
        * d**2
        * np.sqrt(2 * delta_p * rho)
    )

    return qm - qm_calc, beta, e, C, reyn, A


def calc_d(x, state=None, delta_p=None, qm=None, D=None, tappings="flange"):
    r = calc(x, state, delta_p, qm, D, tappings)
    return r[0]



def check_similarity(point_sp, point_t):
    """Function to check similarity between two different points.

    Parameters
    ----------
    point_sp : ccp.Point
        The specified (reference) point in the compressor map.
    point_t : ccp.Point
        The new point in the compressor map.

    Returns
    -------
    similarity_results : dict
        Dict with information on flow, volume ratio, Mach and Reynolds similarity.
    """
    flow_coefficient = point_t.phi / point_sp.phi
    volume_ratio = point_t.volume_ratio / point_sp.volume_ratio
    mach = point_t.mach - point_sp.mach
    reynolds = point_t.reynolds / point_sp.reynolds

    flow_coefficient_limits = (0.96, 1.04)
    volume_ratio_limits = (0.95, 1.05)

    if 0 < point_sp.mach < 0.214:
        mach_limits = (-point_sp.mach.m, -0.25 * point_sp.mach.m + 0.286)
    elif 0.215 < point_sp.mach < 0.86:
        mach_limits = (0.266 * point_sp.mach.m - 0.271, -0.25 * point_sp.mach.m + 0.286)
    elif 0.86 < point_sp.mach:
        mach_limits = (-0.042, 0.07)
    else:
        mach_limits = "Mach outside PTC10 limits."

    x = (point_sp.reynolds / 1e7) ** 0.3
    if 9e4 < point_sp.reynolds < 1e7:
        upper = 100**x
    elif 1e7 < point_sp.reynolds:
        upper = 100
    else:
        upper = "Reynolds outside PTC10 limits."

    if 9e4 < point_sp.reynolds < 1e6:
        lower = 0.01**x
    elif 1e6 < point_sp.reynolds:
        lower = 0.1
    else:
        lower = "Reynolds outside PTC10 limits."

    reynolds_limits = (lower, upper)

    similarity_results = f"""
    {flow_coefficient.m:.3f} Limits -> {flow_coefficient_limits}
    {volume_ratio.m:.3f} Limits -> {volume_ratio_limits}
    {mach.m:.3f} Limits -> {mach_limits}
    {reynolds.m:.3f} Limits -> {reynolds_limits}
    """

    return similarity_results



import ccp
import dash
from dash import dcc
from dash import html
import dash_bootstrap_components as dbc
import numpy as np
import plotly.graph_objs as go
import plotly.io as pio
from pathlib import Path
from dash.dependencies import Input, Output
from dash.exceptions import PreventUpdate
from app_engine import calculate_performance, get_imp_fd

app = dash.Dash(external_stylesheets=[dbc.themes.FLATLY], title="ccp - Dashboard")
app.config.suppress_callback_exceptions = True


DATA_PATH = Path(ccp.__file__).parent / "app/data"
COMPRESSOR_TAGS = ["C-1231-A", "C-1231-B"]  # , "C-1231-C", "C-1231-D", "C-1231-E"]


# the style arguments for the sidebar. We use position:fixed and a fixed width
SIDEBAR_STYLE = {
    "position": "fixed",
    "top": 0,
    "left": 0,
    "bottom": 0,
    "width": "16rem",
    "padding": "2rem 1rem",
    "background-color": "#f8f9fa",
}

# the styles for the main content position it to the right of the sidebar and
# add some padding.
CONTENT_STYLE = {
    "margin-left": "18rem",
    "margin-right": "2rem",
    "padding": "2rem 1rem",
}

sidebar = html.Div(
    [
        html.H1(
            [
                html.Img(
                    src="/assets/ccp.png", width=150, style={"display": "inline-block"}
                ),
                html.P("ccp Dashboard", className="lead", style={"font-size": 26}),
            ],
            style={"text-align": "center"},
            className="display-4",
        ),
        html.Hr(),
        html.P(
            "Monitoramento de Performance dos Compressores da UTGCA",
            className="lead",
            style={"fontSize": "0.8rem"},
        ),
        dbc.Nav(
            [
                dbc.NavLink(f"{ctag}", href=f"/{ctag}", active="exact")
                for ctag in COMPRESSOR_TAGS
            ],
            vertical=True,
            pills=True,
        ),
    ],
    style=SIDEBAR_STYLE,
)

content = html.Div(id="page-content", style=CONTENT_STYLE)
interval = dcc.Interval(
    id="interval-component",
    interval=900 * 1000,
    n_intervals=0,  # in milliseconds
)
store = dcc.Store(id="store")
store_int = dcc.Store(id="store-int")


app.layout = html.Div(
    [dcc.Location(id="url"), sidebar, content, store, interval, store_int]
)


@app.callback(Output("page-content", "children"), [Input("url", "pathname")])
def render_page_content(pathname):
    if pathname[1:] in COMPRESSOR_TAGS:
        return page_layout()
    # If the user tries to reach a different page, return a 404 message
    return dbc.Jumbotron(
        [
            html.H1("404: Not found", className="text-danger"),
            html.Hr(),
            html.P(f"The pathname {pathname} was not recognised..."),
        ]
    )


def page_layout():
    page = [
        dbc.Tabs(
            [
                dbc.Tab(label="Atual", tab_id="atual"),
                dbc.Tab(label="Intervalo", tab_id="intervalo"),
                dbc.Tab(label="Trend", tab_id="trend"),
            ],
            id="tabs",
            active_tab="atual",
        ),
        html.Div(id="tab-content", className="p-4"),
    ]
    return page


@app.callback(
    Output("tab-content", "children"),
    [
        Input("tabs", "active_tab"),
        Input("store", "data"),
        Input("store-int", "data"),
        Input("url", "pathname"),
    ],
)
def render_tab_content(active_tab, data, data_int, pathname):
    """
    This callback takes the 'active_tab' property as input, as well as the
    stored graphs, and renders the tab content depending on what the value of
    'active_tab' is.
    """
    path = pathname.replace("/", "")
    if active_tab and data is not None:
        if active_tab == "atual":
            # TODO Insert sample time
            container = dbc.Container(
                [
                    dbc.Row(
                        [
                            dbc.Col(html.P("Horário da medição: "), width=4),
                            dbc.Col(html.P(data[f"sample-time-{path}"]), width=4),
                        ]
                    ),
                    dbc.Row(dbc.Col(dcc.Graph(figure=data[f"head-{path}"]), width=12)),
                    dbc.Row(dbc.Col(dcc.Graph(figure=data[f"eff-{path}"]), width=12)),
                    dbc.Row(dbc.Col(dcc.Graph(figure=data[f"power-{path}"]), width=12)),
                ],
            )
            return container
        elif active_tab == "intervalo":

            if data_int is None:
                data_int = data
            else:
                if f"head-{path}" not in data_int:
                    data_int = data
            container = dbc.Container(
                [
                    dbc.Row(
                        [
                            dbc.Col(
                                dbc.FormGroup(
                                    [
                                        dbc.Input(
                                            id="start-date",
                                            persistence=True,
                                            type="text",
                                        ),
                                        dbc.FormText("Data e Horário"),
                                        dbc.Popover(
                                            dbc.PopoverBody(
                                                "Data e Horário em formato aceito pelo PI."
                                            ),
                                            id="popover-start-date",
                                            target="start-date",
                                            trigger="hover",
                                            placement="bottom",
                                        ),
                                    ]
                                ),
                                width=6,
                            ),
                            dbc.Col(
                                dbc.Button(
                                    "Calcular",
                                    color="primary",
                                    block=True,
                                    id="calcular",
                                ),
                                width=6,
                            ),
                        ]
                    ),
                    dbc.Row(
                        dbc.Col(dcc.Graph(figure=data_int[f"head-{path}"]), width=12)
                    ),
                    dbc.Row(
                        dbc.Col(dcc.Graph(figure=data_int[f"eff-{path}"]), width=12)
                    ),
                    dbc.Row(
                        dbc.Col(dcc.Graph(figure=data_int[f"power-{path}"]), width=12)
                    ),
                ],
            )
            return container
        elif active_tab == "trend":
            return dbc.Row(
                [
                    dbc.Col(dcc.Graph(figure=data["hist_1"]), width=6),
                    dbc.Col(dcc.Graph(figure=data["hist_2"]), width=6),
                ]
            )
    return "No tab selected"


@app.callback(Output("store", "data"), [Input("interval-component", "n_intervals")])
def generate_graphs(n_intervals):
    """
    This callback generates three simple graphs from random data.
    """
    # simulate expensive graph generation process
    print("running")
    print("n_intervals: ", n_intervals)

    results_dict = {}

    for ctag in COMPRESSOR_TAGS:
        print(f"Loading {ctag}.")
        json_files = [str(f) for f in DATA_PATH.glob("*.json")]
        json_files = [f for f in json_files if ctag in f]
        json_files = sorted(json_files)

        eff_file = json_files[0]
        sample_datetime_raw = eff_file.split("time-")[-1]
        sample_date, sample_time = sample_datetime_raw.split("_")
        sample_time = sample_time.replace("-", ":").replace(".json", "")

        sample_datetime = sample_date + " " + sample_time

        results_dict[f"sample-time-{ctag}"] = sample_datetime

        for curve in ["head", "eff", "power"]:
            # Head
            file_path = DATA_PATH / f"{ctag}_{curve}_plot-time-{sample_datetime_raw}"
            fig = pio.read_json(str(file_path))
            results_dict[f"{curve}-{ctag}"] = fig

    # save figures in a dictionary for sending to the dcc.Store
    return results_dict


@app.callback(
    Output("store-int", "data"),
    Output("calcular", "n_clicks"),
    Input("calcular", "n_clicks"),
    Input("start-date", "value"),
    Input("url", "pathname"),
    Input("store-int", "data"),
    prevent_initial_call=True,
)
def generate_interval_fig(n_clicks, start_date, pathname, data):
    """
    This callback generates three simple graphs from random data.
    """
    # simulate expensive graph generation process
    print("running calc")
    print("n_clicks: ", n_clicks)
    print(data)

    results_dict = {}

    if n_clicks in [0, None]:
        return data, 0
    else:
        ctag = pathname.replace("/", "")

        print(f"Calculating {ctag} interval.")
        imp_op, point_op, sample_time = calculate_performance(
            ctag.split("-")[-1].lower(), 10, imp_fd
        )

        results_dict[f"sample-time-{ctag}"] = sample_time

        for curve in ["head", "eff", "power"]:
            # Head
            fig = getattr(imp_op, f"{curve}_plot")(
                flow_v=point_op.flow_v, speed=point_op.speed, speed_units="RPM"
            )
            fig = getattr(point_op, f"{curve}_plot")(
                fig=fig, speed_units="RPM", name="Operation Point"
            )

            fig.for_each_trace(
                lambda trace: trace.update(name="Expected Point")
                if "Flow" in trace.name
                else ()
            )
            results_dict[f"{curve}-{ctag}"] = fig

        print(f"Calculating {ctag} interval ended.")
        n_clicks = 0

        # save figures in a dictionary for sending to the dcc.Store
    print("n_clicks: ", n_clicks)
    return results_dict, n_clicks


if __name__ == "__main__":
    imp_fd = get_imp_fd()
    app.run_server(host="0.0.0.0", port=8080, debug=False)



import pytest
from pathlib import Path
from numpy.testing import assert_allclose
import ccp

Q_ = ccp.Q_

test_dir = Path(__file__).parent
data_dir = test_dir / "data"


@pytest.fixture
def impeller_lp_sec1_from_csv():
    suc = ccp.State.define(
        p=Q_(4.08, "bar"),
        T=Q_(33.6, "degC"),
        fluid={
            "METHANE": 58.976,
            "ETHANE": 3.099,
            "PROPANE": 0.6,
            "N-BUTANE": 0.08,
            "I-BUTANE": 0.05,
            "N-PENTANE": 0.01,
            "I-PENTANE": 0.01,
            "NITROGEN": 0.55,
            "HYDROGEN SULFIDE": 0.02,
            "CARBON DIOXIDE": 36.605,
        },
    )
    imp = ccp.Impeller.load_from_engauge_csv(
        suc=suc,
        curve_name="lp-sec1-caso-a",
        curve_path=data_dir,
        b=Q_(5.7, "mm"),
        D=Q_(550, "mm"),
        head_units="kJ/kg",
        flow_units="m³/h",
        speed_units="RPM",
        number_of_points=7,
    )

    return imp


def test_impeller_lp_sec1_from_csv(impeller_lp_sec1_from_csv):
    p0 = impeller_lp_sec1_from_csv[0]
    assert_allclose(p0.flow_v.to("m³/h"), 11250)
    assert_allclose(p0.speed.to("RPM"), 6882.0)
    assert_allclose(p0.disch.p().to("bar"), 9.02198284)
    assert_allclose(p0.head.to("kJ/kg"), 82.7824137)
    assert_allclose(p0.eff, 0.789412, rtol=1e-5)
    assert_allclose(p0.power.to("kW"), 1431.03994)



import io
import os
import re

from setuptools import setup, find_packages


def read(path, encoding="utf-8"):
    path = os.path.join(os.path.dirname(__file__), path)
    with io.open(path, encoding=encoding) as fp:
        return fp.read()


def version(path):
    """Obtain the package version from a python file e.g. pkg/__init__.py
    See <https://packaging.python.org/en/latest/single_source_version.html>.
    """
    version_file = read(path)
    version_match = re.search(
        r"""^__version__ = ['"]([^'"]*)['"]""", version_file, re.M
    )
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


DESCRIPTION = "Centrifugal Compressor Performance calculation."
VERSION = version("ccp/__init__.py")

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION


setup(
    name="ccp-performance",
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Raphael Timbó",
    author_email="raphaelts@gmail.com",
    package_data={"ccp.config": ["new_units.txt"], "ccp.tests.data": ["*"]},
    python_requires=">=3.6",
    install_requires=[
        "numpy",
        "scipy",
        "CoolProp",
        "pint>=0.18",
        "plotly",
        "toml",
        "openpyxl",
        "tqdm",
        "ctREFPROP",
    ],
    extras_require={
        "dev": [
            "pytest",
            "pytest-cov",
            "coverage",
            "codecov",
            "sphinx",
            "nbsphinx",
            "numpydoc",
            "sphinxcontrib-bibtex",
            "black",
            "isort",
            "myst-nb",
            "linkify-it-py",
            "sphinx-book-theme",
            "sphinx-panels",
            "sphinx-copybutton",
            "sphinx-rtd-theme",
        ],
    },
    license="MIT",
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: Implementation :: CPython",
        "Operating System :: OS Independent",
    ],
)



import csv

import numpy as np
import toml
from scipy.interpolate import interp1d
import plotly.graph_objects as go

from ccp import Q_, ureg, Point


class StateParameter:
    def __init__(self, curve_state_object, attr):
        self.curve_state_object = curve_state_object
        self.attr = attr

    def __call__(self, *args, **kwargs):
        values = []

        for point in self.curve_state_object:
            values.append(getattr(getattr(point, self.attr)(), "magnitude"))

        units = getattr(getattr(point, self.attr)(), "units")

        return Q_(values, units)


def state_parameter(curve_state_object, attr):
    return StateParameter(curve_state_object, attr)


class PlotFunction:
    def __init__(self, curve_state_object, attr):
        self.curve_state_object = curve_state_object
        self.attr = attr

    def __call__(self, *args, plot_kws=None, **kwargs):
        """Plot parameter versus volumetric flow.

        You can choose units with the arguments flow_v_units='...' and
        {attr}_units='...'. For the speed you can use speed_units='...'.
        """
        curve_state_object = self.curve_state_object
        attr = self.attr
        fig = kwargs.pop("fig", None)

        if fig is None:
            fig = go.Figure()

        if plot_kws is None:
            plot_kws = {}

        x_units = kwargs.get("flow_v_units", curve_state_object.flow_v.units)
        x_units = ureg.Unit(x_units)
        try:
            y_units = kwargs.get(
                f"{attr}_units", getattr(curve_state_object, attr).units
            )
        except AttributeError:
            y_units = kwargs.get(
                f"{attr}_units", getattr(curve_state_object, attr)().units
            )
        y_units = ureg.Unit(y_units)
        speed_units = kwargs.get("speed_units", curve_state_object.speed.units)
        speed_units = ureg.Unit(speed_units)
        name = kwargs.get("name", str(round(curve_state_object.speed.to(speed_units))))

        flow_v = curve_state_object.flow_v

        interpolated_curve = getattr(curve_state_object, attr + "_interpolated")

        flow_v_range = np.linspace(min(flow_v), max(flow_v), 30)

        values_range = interpolated_curve(flow_v_range)

        values_range = values_range.magnitude

        if x_units is not None:
            flow_v_range = (
                Q_(flow_v_range, curve_state_object.flow_v.units).to(x_units).m
            )
        if y_units is not None:
            try:
                values_range = (
                    Q_(values_range, getattr(curve_state_object, attr).units)
                    .to(y_units)
                    .m
                )
            except AttributeError:
                values_range = (
                    Q_(values_range, getattr(curve_state_object, attr)().units)
                    .to(y_units)
                    .m
                )

        fig.add_trace(go.Scatter(x=flow_v_range, y=values_range, name=name), **plot_kws)

        fig.update_layout(
            xaxis=dict(title=f"Volume Flow ({x_units:~H})"),
            yaxis=dict(title=f"{attr.capitalize()} ({y_units:~H})"),
        )

        return fig


def plot_func(curve_state_object, attr):
    return PlotFunction(curve_state_object, attr)


class InterpolatedFunction:
    def __init__(self, curve_state_object, attr):
        self.curve_state_object = curve_state_object
        self.attr = attr

    def __call__(self, *args, **kwargs):
        curve_state_object = self.curve_state_object
        attr = self.attr
        values = getattr(curve_state_object, attr)
        if callable(values):
            values = values()

        units = values.units

        if len(values) < 3:
            interpolation_degree = 1
        else:
            interpolation_degree = 3

        interpol_function = interp1d(
            curve_state_object.flow_v.magnitude,
            values.magnitude,
            kind=interpolation_degree,
            fill_value="extrapolate",
        )

        try:
            args = [arg.magnitude for arg in args]
        except AttributeError:
            pass

        result = Q_(interpol_function(*args, **kwargs), units)
        if isinstance(*args, (int, float)):
            result = Q_(float(result.magnitude), result.units)

        return result


def interpolated_function(curve_state_object, attr):
    return InterpolatedFunction(curve_state_object, attr)


class _CurveState:
    """Class used to create list with states from curve.

    This enables the following call:
    # >>> curve.suc.p()
    (100000, 100000) pascal

    """

    def __init__(self, points, flow_v, speed):
        self.flow_v = flow_v
        self.points = points
        self.speed = speed

        # set a method for each suction attribute in the list
        for attr in ["p", "T", "h", "s", "rho"]:
            func = state_parameter(self, attr)
            setattr(self, attr, func)

            interpol_func = interpolated_function(self, attr)
            setattr(self, f"{attr}_interpolated", interpol_func)

            plot = plot_func(self, attr)
            setattr(self, f"{attr}_plot", plot)

    def __getitem__(self, item):
        return self.points.__getitem__(item)


class Curve:
    """Curve.

    A curve is a collection of points that share the same suction
    state and the same speed.

    Parameters
    ----------

    points : list
        List with the points
    """

    def __init__(self, points):
        if len(points) < 2:
            raise TypeError("At least 2 points should be given.")
        self.points = sorted(points, key=lambda p: p.flow_v)

        _flow_v_values = [p.flow_v.magnitude for p in self]
        _flow_v_units = self[0].flow_v.units
        self.flow_v = Q_(_flow_v_values, _flow_v_units)

        self.speed = self[0].speed
        # change the following check in the future
        for point in self:
            if self.speed != point.speed:
                raise ValueError("Speed for each point should be equal")

        self.suc = _CurveState(
            [p.suc for p in self], flow_v=self.flow_v, speed=self.speed
        )
        self.disch = _CurveState(
            [p.disch for p in self], flow_v=self.flow_v, speed=self.speed
        )

        for param in ["head", "eff", "power", "phi", "psi"]:
            values = []
            for point in self:
                try:
                    values.append(getattr(getattr(point, param), "magnitude"))
                    units = getattr(getattr(point, param), "units")
                except AttributeError:
                    continue

            setattr(self, param, Q_(values, units))

            interpol_func = interpolated_function(self, param)
            setattr(self, f"{param}_interpolated", interpol_func)

            plot = plot_func(self, param)
            setattr(self, param + "_plot", plot)

    def __getitem__(self, item):
        return self.points.__getitem__(item)

    def __iter__(self):
        for point in self.points:
            yield point

    def __len__(self):
        return len(self.points)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            for point_self, point_other in zip(self.points, other.points):
                if point_self != point_other:
                    return False
            else:
                return True

    def _dict_to_save(self):
        return {f"point{i}": point._dict_to_save() for i, point in enumerate(self)}

    def save(self, file_name, file_type="toml"):
        """Save curve to a file.

        Parameters
        ----------
        file_name: str
            Name of the file.
        file_type: str
            File type can be: toml.
        """
        if file_type == "toml":
            with open(file_name, mode="w") as f:
                toml.dump(self._dict_to_save(), f)

    def save_hysys_csv(self, curve_path):
        """Save curve to a csv with hysys format.

        curve_path: pathlib.Path
            Path where the curve file will be saved.
        """
        curve_path.parent.mkdir(parents=True, exist_ok=True)
        with open(str(curve_path), mode="w") as csv_file:
            field_names = ["Volume Flow (m3/h)", "Head (m)", "Efficiency (%)"]
            writer = csv.DictWriter(csv_file, fieldnames=field_names)
            writer.writeheader()
            for point in self:
                writer.writerow(
                    {
                        "Volume Flow (m3/h)": point.flow_v.to("m**3/h").m,
                        "Head (m)": point.head.m / 9.81,
                        "Efficiency (%)": 100 * point.eff.m,
                    }
                )

    @classmethod
    def load(cls, file_name):
        with open(file_name) as f:
            parameters = toml.load(f)

        return cls(
            [Point(**Point._dict_from_load(kwargs)) for kwargs in parameters.values()]
        )
